import {
  object, string, number, date, ValidationError, boolean,
} from 'yup';
import type { Request, Response } from 'express';
import prisma from '../client';

/**
 * @todo 1. Implement endpoints for listing animals, start with basic list
 * @see https://github.com/prisma/prisma-examples/tree/latest/typescript/rest-express
 * Requirement: Browse animals
 *
 *
 * @todo 2. Once you have endpoint for getting animals output only one's that aren't adopted
 * @see https://www.prisma.io/docs/concepts/components/prisma-client/filtering-and-sorting
 * Requirement: Browse animals
 *
 *
 * @todo 3. Create filter for isRequested, isAdopted and type also implement limit which is by
 *          default 10 animals.
 * @see https://www.prisma.io/docs/reference/api-reference/prisma-client-reference#findmany
 * Requirement: Filter animals by type
 *
 *
 * As you can see there are many filters that can be implemented
 */
export const list = async (req: Request, res: Response) => {
  const isRequested = req.query['isRequested'] || false;
  const limit = +(req.query['limit'] || 10);

  const type = `${req.query['type']}`;
  const types = ['dog', 'cat'];

  const filterRequested = isRequested ? { not: null } : {};
  const filterType = types.includes(type) ? type : {};

  const animals = await prisma.animal.findMany({
    where: {
      acceptedRequestId: filterRequested,
      adoptedAt: null,
      ...filterType,
    },
    take: limit,
  });

  return res.send({
    status: 'sucess',
    data: animals,
  });
};

/**
 * @todo Create schema for validating animal object before inserting into
 * database, play around different validators, for type use matches and regex.
 * Feel free to also include some defaults.
 * @see https://github.com/jquense/yup
 */
const animalSchema = object({
  name: string().required(),
  description: string().min(10).default('Popis bude pridaný'),
  age: number().positive().required(),
  picture: string().default('/images/animal.webp'),
  type: string().matches(/(cat|dog)/).default('dog'),
  sex: boolean().required(),
  addedAt: date().required(),
});

/**
 * @todo 1. Add animal into database
 * @see https://www.prisma.io/docs/concepts/components/prisma-client/crud#create
 * Requirement: Add animal
 *
 * @todo 2. Check the animal before inserting into database and return correct http status
 *       By checking the instance of error for ValidationError send correct status code
 *       otherwise send generic error.
 * @see https://http.cat/
 */
export const store = async (req: Request, res: Response) => {
  try {
    const data = await animalSchema.validate(req.body);
    const animal = await prisma.animal.create({
      data,
    });

    return res.status(201).send({
      status: 'success',
      data: animal,
      message: 'Animal stored in system',
    });
  } catch (e) {
    if (e instanceof ValidationError) {
      return res.status(400).send({
        status: 'error',
        data: e.errors,
        message: e.message,
      });
    }

    return res.status(500).send({
      status: 'error',
      data: {},
      message: 'Something went wrong',
    });
  }
};

/**
 * Optional
 *
 * @todo Find animal by id
 * @see https://www.prisma.io/docs/reference/api-reference/prisma-client-reference#findunique
 *
 * @todo If no animal is found then return 404 and corresponding message with empty data
 * @see https://expressjs.com/en/guide/error-handling.html
 *
 * You might ask why do we want to return 404 on single animal and not on list?
 * Well, we are expecting a single result, which should contain some data, the list
 * is returned but it's empty because the server might not found the data that matched
 * our criteria.
 */
export const get = async (req: Request, res: Response) => {
  const { id } = req.params;

  if (!id) {
    return res.status(400).send({
      status: 'error',
      data: {},
      message: 'No id was provided',
    });
  }

  const animal = await prisma.animal.findUnique({
    where: {
      id,
    },
  });

  if (!animal) {
    return res.status(404).send({
      status: 'error',
      data: {},
      message: 'Animal was not found',
    });
  }

  return res.send({
    status: 'sucess',
    data: animal,
  });
};
