import type { Request, Response } from 'express';
import {
  object, string, boolean, ValidationError,
} from 'yup';
import prisma from '../client';

/**
 * Request schema
 */
const requestSchema = object({
  from: string().required(),
  isAccepted: boolean().default(false),
  animalId: string().required().uuid(),
});

/**
 * @todo If the request from client contains isAccepted then you need to mark animal
 *       as adopted and fill all required fields on both Request and Animal and update them.
 * @see https://stackoverflow.com/questions/28459418/use-of-put-vs-patch-methods-in-rest-api-real-life-scenarios
 */
export const partialUpdate = async (req: Request, res: Response) => {
  const { id } = req.params;

  if (!id) {
    return res.status(500).send(
      {
        status: 'error',
        message: 'Failed - no provided id',
      },
    );
  }

  try {
    const updatedRequest = await prisma.$transaction(async (tx) => {
      const request = await tx.request.findUniqueOrThrow({
        where: {
          id,
        },
      });

      await tx.animal.update({
        data: {
          acceptedRequest: {
            connect: {
              id,
            },
          },
          adoptedAt: new Date(),
        },
        where: {
          id: request.animalId,
        },
      });

      return tx.request.update({
        data: {
          isAccepted: true,
        },
        where: {
          id,
        },
      });
    });

    return res.send({
      status: 'sucess',
      data: updatedRequest,
      message: 'Request updated',
    });
  } catch (e) {
    return res.status(500).send(
      {
        status: 'error',
        message: 'Something went wrong',
      },
    );
  }
};

/**
 * Update all fields of request
 */
export const update = async (req: Request, res: Response) => {
  const { id } = req.params;

  if (!id) {
    return res.status(500).send(
      {
        status: 'failure',
        message: 'Failed - no provided id',
      },
    );
  }

  try {
    const requestData = await requestSchema.validate(req.body);
    const request = await prisma.request.update({
      data: requestData,
      where: {
        id,
      },
    });

    return res.status(200).send({
      status: 'success',
      data: request,
      message: 'Request updated',
    });
  } catch (e) {
    if (e instanceof ValidationError) {
      return res.status(400).send({
        status: 'error',
        data: e.errors,
        message: e.message,
      });
    }

    return res.status(500).send({
      status: 'error',
      data: {},
      message: 'Something went wrong',
    });
  }
};

/**
 * Store new request
 * Requirement: Request adoption
 */
export const store = async (req: Request, res: Response) => {
  try {
    const data = await requestSchema.validate(req.body);
    const request = await prisma.request.create({
      data,
    });

    return res.status(201).send({
      status: 'success',
      data: request,
      message: 'Request stored in system',
    });
  } catch (e) {
    if (e instanceof ValidationError) {
      return res.status(400).send({
        status: 'error',
        data: e.errors,
        message: e.message,
      });
    }

    return res.status(500).send({
      status: 'error',
      data: {},
      message: 'Something went wrong',
    });
  }
};

/**
 * Show all requests
 */
export const list = async (_req: Request, res: Response) => {
  const requests = await prisma.request.findMany({
    where: {
      deletedAt: null,
    },
  });

  return res.send({
    status: 'sucess',
    data: requests,
  });
};

/**
 * Show single request
 */
export const show = async (req: Request, res: Response) => {
  const { id } = req.params;

  if (!id) {
    return res.status(500).send({
      status: "error",
      data: {},
      message: "Failed - no id provided"
    });
  }
  const request = await prisma.request.findUnique({
    where: {
      id,
    },
  });

  return res.send({
    status: 'sucess',
    data: request,
  });
};

/**
 * Optional
 *
 * @todo Delete specific request, delete only sets deletedAt
 */
export const destroy = async (req: Request, res: Response) => {
  const { id } = req.params;

  if (!id) {
    return res.status(500).send({
      status: "error",
      data: {},
      message: "Failed - no id provided"
    });
  }

  const request = await prisma.request.update({
    data: {
      deletedAt: new Date(),
    },
    where: {
      id,
    },
  });

  return res.status(200).send({
    status: 'sucess',
    data: request,
    message: 'Request removed',
  });
};
